from functools import partial
from multiprocessing import Event as event
from subprocess import Popen, TimeoutExpired, STDOUT, PIPE
from typing import List, Optional, Iterator

from boba_fetch.log import log


class SafeProcess:
    def __init__(self, cmd: List[str]) -> None:
        self.cmd = cmd
        self.proc: Optional[Popen] = None
        self.started = event()

    def run(self) -> Iterator[str]:
        with Popen(self.cmd, stderr=STDOUT, stdout=PIPE) as self.proc:
            self.started.set()
            while self.proc.stdout is not None and self.proc.poll() is None:
                line = []
                for byte in iter(partial(self.proc.stdout.read, 1), b''):
                    if byte not in {b'\n', b'\r', b''}:
                        try:
                            line.append(byte.decode())
                        except UnicodeDecodeError:
                            log.debug(f'Unable to decode: {str(byte)}')
                    else:
                        yield ''.join(line)
                        line = []

    @property
    def returncode(self) -> Optional[int]:
        return self.proc.returncode if self.proc else None

    def terminate(self, timeout: int = 10) -> None:
        if self.proc is not None:
            try:
                self.proc.terminate()
                self.proc.wait(timeout)
            except TimeoutExpired:
                self.proc.kill()



from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing.synchronize import Event
from multiprocessing import Event as event, Queue
from os import access, R_OK, W_OK
from pathlib import Path
from queue import Empty
import shlex
from subprocess import run, STDOUT, PIPE, SubprocessError
from threading import Thread
from traceback import format_exc
from typing import Optional, List, ContextManager

from boba_fetch.log import log, thread_name
from boba_fetch.process import SafeProcess
from boba_fetch.rpc import UnixSocketRPCClient


@dataclass(frozen=True)
class WatchParams:
    host: str
    local_dir: Path
    queue: Queue
    delete: bool
    execute_on_complete: Optional[str] = None


@dataclass(frozen=True)
class SyncParams:
    host: str
    local_dir: Path
    remote_path: Path
    delete: bool
    execute_on_complete: Optional[str] = None

    @property
    def on_complete(self) -> Optional[List[str]]:
        return shlex.split(self.execute_on_complete) if self.execute_on_complete else None

    @property
    def name(self) -> str:
        return self.remote_path.name

    @staticmethod
    def from_watch(watch_params: WatchParams, remote_path: Path) -> 'SyncParams':
        return SyncParams(watch_params.host, watch_params.local_dir, remote_path,
                          watch_params.delete, watch_params.execute_on_complete)

    @property
    def no_delete(self) -> 'SyncParams':
        return SyncParams(self.host, self.local_dir, self.remote_path, delete=False)


class Syncer:
    def __init__(self, tracker: UnixSocketRPCClient,
                 halt_flag: Optional[Event] = None,
                 temp_dir: Optional[Path] = None):
        self.halt_flag = halt_flag or event()
        self.sleeping = event()
        self.sleeping.set()
        self.tracker = tracker
        self.temp_dir = temp_dir
        if self.temp_dir and not self.temp_dir.is_dir():
            raise ValueError(f'Temporary directory {self.temp_dir} does not exist!')
        if self.temp_dir and not access(self.temp_dir, R_OK | W_OK):
            raise ValueError(f'No read-write access to temporary directory: {self.temp_dir}')

    def _cmd(self, params: SyncParams) -> List[str]:
        result = ['rsync', '-Pprvzy', '--chmod=D755,F644']
        if self.temp_dir:
            result += ['-T', str(self.temp_dir)]
        if params.delete:
            result += ['--remove-source-files']
        return result + [f'{params.host}:"{str(params.remote_path)}"', str(params.local_dir)]

    def cmd(self, params: SyncParams) -> List[str]:
        return self._cmd(params)

    @contextmanager  # type: ignore # mpypy and typeguard are fighting
    def regular_check_in(self, name) -> ContextManager[None]:  # type: ignore
        stop_check_in = event()
        def repeat():
            while not stop_check_in.is_set():
                self.tracker.check_in(name)
                stop_check_in.wait(timeout=1)
        thread = Thread(target=repeat)
        thread.start()
        try:
            yield
        finally:
            stop_check_in.set()
            thread.join()

    def sync_down(self, params: SyncParams) -> bool:
        proc = SafeProcess(self.cmd(params))
        if not self.halt_flag.is_set():
            for line in proc.run():
                log.info(line)
                if self.halt_flag.is_set():
                    proc.terminate(3)
        return proc.returncode == 0

    def halt(self):
        log.info("Halting")
        self.halt_flag.set()

    @staticmethod
    def execute_on_complete(params: SyncParams) -> None:
        if params.execute_on_complete is not None:
            cmd = params.execute_on_complete.replace('%F', str(params.local_dir / params.name))
            log.info(f'Executing: {cmd}')
            try:
                result = run(shlex.split(cmd), check=False, text=True, stdout=PIPE, stderr=STDOUT)
                log.info(result.stdout)
                log.info(f"Exit Code {result.returncode}")
            except SubprocessError:
                log.error(f"Failed to execute on-complete command: {cmd}")

    def watch_queue(self, watch_params: WatchParams):
        while not self.halt_flag.is_set():
            try:
                remote_path = Path(watch_params.queue.get(timeout=0.5))
                self.sleeping.clear()
                params = SyncParams.from_watch(watch_params, remote_path)
                if not self.tracker.currently_working(params.name):
                    with thread_name(params.name), self.regular_check_in(params.name):
                        log.info(f'Syncing down: {remote_path} ~> {watch_params.local_dir}')
                        self.tracker.start(params.name)
                        try:
                            self.sync_down(params.no_delete)
                            if not self.sync_down(params):  # verification
                                log.error(f"Failed to sync down {params.host}:{params.remote_path}")
                                self.tracker.failed(params.name)
                            else:
                                self.tracker.mark_complete(params.name)
                                log.info(f'Sync complete: {params.local_dir / params.name}')
                                self.execute_on_complete(params)
                                log.info('Post-Sync Task complete.')
                        except Exception:  # pylint: disable=W0703
                            log.error(format_exc())
                            log.error(f"Failed to sync down {params.host}:{params.remote_path}")
                            self.tracker.failed(params.name)
                        finally:
                            self.sleeping.set()
            except Empty:
                continue



from pathlib import Path
from time import time
from typing import Optional, Dict, List, Set

from boba_fetch.log import log
from boba_fetch.rpc import UnixSocketRPCServer


class Tracker(UnixSocketRPCServer):
    def __init__(self, directory: Path, check_in_timeout: float = 5.0) -> None:
        self.working: Dict[str, float] = {}
        self.enqueued: Set[str] = set()
        self.directory = directory
        self.timeout = check_in_timeout
        if not self.directory.is_dir():
            self.directory.mkdir(parents=True, exist_ok=True)
        UnixSocketRPCServer.__init__(self)
        self.thread_name = 'Tracker'

    @UnixSocketRPCServer.serve
    def try_enqueue(self, name: str) -> bool:
        if name not in self.enqueued and not self.currently_working(name):
            log.debug(f'The file [{name}] is not enqueued and not currently being worked. Marking as enqueued.')
            self.enqueued.add(name)
            log.debug(f'The set of enqueued items is: [{list(self.enqueued)}]')
            log.debug(f'The set of items currently being worked is: [{self.unfinished()}]')
            return True
        log.debug(f'The file [{name}] is either marked enqueued or currently being worked.')
        return False

    @UnixSocketRPCServer.serve
    def completed_at(self, name: str) -> float:
        if (self.directory / name).is_file():
            contents = (self.directory / name).open().read().strip()
            log.debug(f'The file [{name}] tracking file contents: {contents}')
            return float(contents) if contents.replace('.', '', 1).isdigit() else 0.0
        log.debug(f'The tracking file for [{name}] does not exist')
        return 0.0

    @UnixSocketRPCServer.serve
    def mark_complete(self, name: str, timestamp: Optional[float] = None) -> None:
        log.debug(f'Marking [{name}] complete')
        (self.directory / name).open('w+').write(str(timestamp or time()))
        self.not_working(name)

    @UnixSocketRPCServer.serve
    def newer_than(self, name: str, timestamp: float) -> bool:
        if self.completed_at(name) >= timestamp:
            log.debug(f'[{name}] was completed at {self.completed_at(name)} which is >= {timestamp}')
        else:
            log.debug(f'[{name}] was completed at {self.completed_at(name)} which is < {timestamp}')
        return self.completed_at(name) >= timestamp

    @UnixSocketRPCServer.serve
    def start(self, name: str) -> None:
        self.check_in(name)
        (self.directory / name).open('w+').write('')
        if name in self.enqueued:
            self.enqueued.remove(name)

    @UnixSocketRPCServer.serve
    def failed(self, name: str) -> None:
        log.debug(f'Marking [{name}] as failed, removing from working set, enqueued, and tracking file')
        (self.directory / name).unlink(missing_ok=True)
        self.not_working(name)
        log.debug(f'The set of enqueued items is: [{list(self.enqueued)}]')
        log.debug(f'The set of items currently being worked is: [{self.unfinished()}]')

    def not_working(self, name: str) -> None:
        if name in self.working:
            del self.working[name]
        if name in self.enqueued:
            self.enqueued.remove(name)

    def clean(self) -> None:
        now = time()
        for name in [n for n, c in self.working.items() if now - c >= self.timeout]:
            self.failed(name)

    @UnixSocketRPCServer.serve
    def check_in(self, name: str) -> None:
        log.debug(f'Checking in for [{name}]')
        self.clean()
        self.working[name] = time()

    @UnixSocketRPCServer.serve
    def currently_working(self, name: str) -> bool:
        return name in self.working and time() - self.working[name] <= self.timeout

    @UnixSocketRPCServer.serve
    def unfinished(self) -> List[str]:
        self.clean()
        return list(self.working)



from dataclasses import dataclass
from multiprocessing import Process, Event as event, Queue
from multiprocessing.synchronize import Event
from pathlib import Path
from signal import SIGTERM, SIGINT, signal
from tempfile import mkdtemp
from typing import Optional, List
from warnings import simplefilter

import click
import cryptography
from pkg_resources import get_distribution  # type: ignore

from boba_fetch.log import log, thread_name, set_loglevel, DEBUG, INFO
from boba_fetch.poller import RemotePoller
from boba_fetch.rpc import UnixSocketRPCClient
from boba_fetch.syncer import Syncer, WatchParams
from boba_fetch.tracker import Tracker


simplefilter('ignore', cryptography.utils.CryptographyDeprecationWarning)


@dataclass
class BobaConfig:  # pylint: disable=R0902  # dataclasses should be allowed more instance vars
    remote_host: str
    remote_dir: Path
    local_dir: Path
    staging_dir: Path
    tracking_dir: Path
    num_sync_workers: int
    delete_src: bool
    execute_on_complete: Optional[str]
    polling_delay: float
    ssh_config: Path = Path.home() / '.ssh' / '.config'

    def to_watch_params(self, queue: Queue) -> WatchParams:
        return WatchParams(self.remote_host, self.local_dir, queue, self.delete_src,
                           self.execute_on_complete)


def run_poller(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event):
    with thread_name("Poller"):
        poller = RemotePoller(config.remote_host, tracker, config.ssh_config)
        poller.poll(config.polling_delay, config.remote_dir, queue, halt_flag)


def start_poller(config: BobaConfig,
                 tracker: UnixSocketRPCClient,
                 queue: Queue,
                 halt_flag: Event) -> Process:
    poller_proc = Process(target=run_poller, args=(config, tracker, queue, halt_flag))
    poller_proc.start()
    return poller_proc


def run_syncer(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event,
               index: int):
    with thread_name(f'Syncer {index}'):
        syncer = Syncer(tracker, halt_flag, config.staging_dir)
        syncer.watch_queue(config.to_watch_params(queue))


def start_syncers(config: BobaConfig,
                  tracker: UnixSocketRPCClient,
                  queue: Queue,
                  halt_flag: Event) -> List[Process]:
    syncer_procs = [Process(target=run_syncer,
                            args=(config, tracker,
                                  queue, halt_flag, i)) for i in range(config.num_sync_workers)]
    for proc in syncer_procs:
        proc.start()
    return syncer_procs


@click.command()
@click.version_option(get_distribution('boba-fetch').version)
@click.option('-r', '--remote', required=True,
              help='The remote host to poll for file updates. Must be SSH accessible via public key '
                   'and configured in your ~/.ssh/config')
@click.option('-s', '--src-dir', required=True, type=click.Path(path_type=Path),
              help='The directory on the remote host to check for files to download')
@click.option('-d', '--dst-dir', required=True, type=click.Path(exists=True,
                                                                file_okay=False,
                                                                dir_okay=True,
                                                                writable=True,
                                                                readable=True,
                                                                resolve_path=True,
                                                                path_type=Path),
              help='Local directory to sync the files to from the remote host')
@click.option('-g', '--staging-dir', default=None, type=click.Path(exists=True,
                                                                   file_okay=False,
                                                                   dir_okay=True,
                                                                   writable=True,
                                                                   readable=True,
                                                                   resolve_path=True,
                                                                   allow_dash=False,
                                                                   path_type=Path),
              help='Staging directory where files are saved, defaults to randomized temporary directory')
@click.option('-t', '--tracking-dir', required=True, type=click.Path(exists=True,
                                                                     file_okay=False,
                                                                     dir_okay=True,
                                                                     writable=True,
                                                                     readable=True,
                                                                     resolve_path=True,
                                                                     allow_dash=False,
                                                                     path_type=Path),
              help='Tracking directory where we keep track of files that have been synced')
@click.option('-n', '--num-sync-workers', default=3, type=int,
              help='Number of sync workers to use for downloads, defaults to 3')
@click.option('--delete-src', is_flag=True, default=False, help='Indicates to whether to delete the source files on '
                                                                'the remote system')
@click.option('-e', '--execute-on-complete', default=None, help='A command to execute every time a sync is complete')
@click.option('--ssh-config', default=None, type=click.Path(exists=True,
                                                            file_okay=True,
                                                            dir_okay=False,
                                                            readable=True,
                                                            resolve_path=True,
                                                            allow_dash=False,
                                                            path_type=Path),
              help='Path to an SSH configuration file (note that "Include" directives are not yet supported).')
@click.option('-p', '--polling-delay', default=30.0, type=float,
              help='Number of '  # pylint: disable=R0914  # application entry-point, needs local variables
                   'seconds to wait between polls of the remote directory, defaults to 30')
@click.option('--debug', is_flag=True, default=False, help='Turn on debug logging')
def main(remote: str,
         src_dir: Path,
         dst_dir: Path,
         staging_dir: Optional[Path],
         tracking_dir: Path,
         num_sync_workers: int,
         delete_src: bool,
         execute_on_complete: Optional[str],
         ssh_config: Optional[Path],
         polling_delay: float,
         debug: bool):
    set_loglevel(DEBUG if debug else INFO)
    queue: Queue = Queue()
    halt_flag = event()
    config = BobaConfig(remote, src_dir, dst_dir, staging_dir or Path(mkdtemp()),
                        tracking_dir, num_sync_workers, delete_src, execute_on_complete,
                        polling_delay, ssh_config)
    with Tracker(config.tracking_dir).proxy() as tracker:  # type: ignore
        syncer_procs = start_syncers(config, tracker, queue, halt_flag)
        poller_proc = start_poller(config, tracker, queue, halt_flag)

        def handle_signal(_, __):
            halt_flag.set()
        signal(SIGTERM, handle_signal)
        signal(SIGINT, handle_signal)
        poller_proc.join()
        log.info("Poller shutdown complete.")
        for sync_proc in syncer_procs:
            sync_proc.join()
        log.info("Syncer shutdown complete.")
        log.info("Done.")



from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator


@contextmanager
def tempdir() -> Generator[Path, None, None]:
    with TemporaryDirectory() as tmp:
        yield Path(tmp)


def generate_big_sparse_file(path: Path, size) -> Path:
    with path.open('wb+') as file:
        file.seek(size - 1)
        file.write(b"\1")
        file.close()
    return path



from time import time, sleep

from boba_fetch.tracker import Tracker

from tests.unit.utils import tempdir


def test_tracker_basics() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            assert tracker.completed_at('some file.txt') == 0.0


def test_tracker_directory_creation() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            assert tracker.completed_at("test.txt") == 0.0


def test_mark_complete() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            start = time()
            tracker.mark_complete('test.txt')
            assert tracker.completed_at('test.txt') >= start


def test_start() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            tracker.start('test')
            assert (tmp_dir / 'test').is_file()


def test_unfinished() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            tracker.start('test')
            assert tracker.unfinished() == ['test']


def test_failed() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            tracker.start('test')
            assert (tmp_dir / 'test').is_file()
            assert tracker.unfinished() == ['test']
            tracker.failed('test')
            assert not (tmp_dir / 'test').is_file()
            assert not tracker.unfinished()


def test_try_enqueue() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            assert tracker.try_enqueue('test')
            assert not tracker.try_enqueue('test')


def test_try_enqueue_and_start() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            assert tracker.try_enqueue('test')
            tracker.start('test')
            assert not tracker.try_enqueue('test')


def test_start_and_try_enqueue() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            tracker.start('test')
            assert not tracker.try_enqueue('test')


def test_newer_than() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir).proxy() as tracker:
            tracker.mark_complete('test', 16000000)
            assert tracker.newer_than('test', 15000000)


def test_check_in_expiration() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir, check_in_timeout=0.2).proxy() as tracker:
            tracker.start('test')
            assert tracker.currently_working('test')
            assert tracker.unfinished() == ['test']
            sleep(0.2)
            assert tracker.unfinished() == []
            assert not tracker.currently_working('test')
